@@ -11,6 +11,10 @@ This file is licensed to you under the license specified in the included file
 #include "config.h"
 #endif
 
+#include <list>
+
+#include "acceptedvalues.h"
+
 struct gengetopt_option
 {
   char short_opt;
@@ -19,7 +23,6 @@ struct gengetopt_option
   int type; /* values in `argsdef.h' */
   int flagstat ;
   int required;
-  struct gengetopt_option * next;
   char * var_arg; /* canonized long_opt + "_arg" = argument var */
   int default_given ; /* if a default is given */
   char * default_string ; /* default value for this option, if string */
@@ -29,8 +32,11 @@ struct gengetopt_option
   bool arg_is_optional; /* whether the argument is optional */
   char *type_str; /* Alternative name for type,
                      e.g. "URL" or "SECONDS" */
+  const AcceptedValues *acceptedvalues; /* values that can be passed to this option */
   char *section; /* the section of this option */
   char *section_desc; /* the description associated with the possible section */
 };
 
+typedef std::list<gengetopt_option *> gengetopt_option_list;
+
 #endif /* _GENGETOPT_GGOS_H */
