@@ -41,13 +41,8 @@ extern "C"
 #include "my_sstream.h"
 
 #include "groups.h"
-#include "string_opt_arg.h"
-#include "string_opt_arg_no_free.h"
-#include "int_opt_arg.h"
-#include "long_opt_arg.h"
-#include "long_long_opt_arg.h"
-#include "float_opt_arg.h"
-#include "flag_opt_arg.h"
+#include "update_arg.h"
+#include "option_arg.h"
 #include "required_option.h"
 #include "generic_option.h"
 #include "group_option.h"
@@ -55,13 +50,10 @@ extern "C"
 #include "generic_option_group.h"
 #include "handle_help.h"
 #include "handle_version.h"
-#include "no_short_option.h"
 #include "print_help_string.h"
 #include "multiple_opt_list.h"
 #include "multiple_opt_struct.h"
 #include "multiple_option.h"
-#include "multiple_option_no_short.h"
-#include "multiple_option_new_list_node.h"
 #include "multiple_fill_array.h"
 #include "multiple_fill_array_default.h"
 #include "free_string.h"
@@ -70,6 +62,9 @@ extern "C"
 #include "reset_group.h"
 #include "exit_failure.h"
 #include "update_given.h"
+#include "given_field.h"
+#include "clear_given.h"
+#include "free_list.h"
 #include "gm_utils.h"
 #include "fileutils.h"
 
@@ -79,6 +74,8 @@ extern "C" char *strdup (const char *s) ;
 
 #define MAX_STARTING_COLUMN 32
 
+#define EXE_NAME "argv[0]"
+
 using std::endl;
 
 extern char * gengetopt_package;
@@ -95,7 +92,8 @@ static void _generate_option_arg(ostream &stream,
 static void
 generate_help_desc_print(ostream &stream,
                          unsigned int desc_column,
-                         const char *descript, const char *defval);
+                         const char *descript, const char *defval,
+                         const string &values);
 
 CmdlineParserCreator::CmdlineParserCreator (char *function_name,
                                             char *struct_name,
@@ -166,20 +164,12 @@ CmdlineParserCreator::CmdlineParserCreator (char *function_name,
   c_source_gen_class::set_parser_name (parser_function_name);
   set_source_name (filename);
 
-  if (! no_handle_error)
-    {
-      ostringstream exit_failure_str;
-      exit_failure_gen_class exit_gen;
-      exit_gen.set_parser_name (c_source_gen_class::parser_name);
-
-      exit_gen.generate_exit_failure (exit_failure_str);
-
-      set_final_exit (exit_failure_str.str ());
-    }
-  else
-    set_final_exit ("return result;");
-
-  set_gen_exit (no_handle_error ? "return" : "exit");
+  ostringstream exit_failure_str;
+  exit_failure_gen_class exit_gen;
+  exit_gen.set_parser_name (c_source_gen_class::parser_name);
+  exit_gen.set_handle_error (! no_handle_error);
+  exit_gen.generate_exit_failure (exit_failure_str);
+  set_final_exit (exit_failure_str.str ());
 
   set_conf_parser (conf_parser);
 
@@ -196,9 +186,14 @@ CmdlineParserCreator::CmdlineParserCreator (char *function_name,
     }
 
   set_do_generate_strdup(gen_strdup);
-  set_multiple_token_functions(has_multiple_options());
-  set_multiple_token_vars(has_multiple_options());
+  set_check_possible_values(has_values());
+  set_multiple_token_functions(has_multiple_options_with_type());
+  set_multiple_token_vars(has_multiple_options_with_type());
   set_handle_unamed(unamed_options);
+  set_check_required_options(has_required());
+  set_has_purpose((gengetopt_purpose != 0));
+  set_purpose(generate_purpose());
+  set_no_package((gengetopt_package == 0));
 }
 
 void
@@ -217,50 +212,39 @@ CmdlineParserCreator::do_update_arg (struct gengetopt_option *opt,
                                      unsigned int indent,
                                      const string argstr = "optarg" )
 {
+  if (opt->type == ARG_NO)
+    {
+      stream << "break;";
+      return;
+    }
+  
   string indent_str (indent, ' ');
   const int offset = 4;
 
-  // FIXME: do not create all these generators for a single option
-  string_opt_arg_gen_class str_gen;
-  str_gen.set_optarg (argstr);
-  string_opt_arg_no_free_gen_class str_gen_no_free;
-  str_gen_no_free.set_optarg (argstr);
-  int_opt_arg_gen_class int_gen;
-  int_gen.set_optarg (argstr);
-  long_opt_arg_gen_class long_gen;
-  long_gen.set_optarg (argstr);
-  long_long_opt_arg_gen_class long_long_gen;
-  long_long_gen.set_optarg (argstr);
-  float_opt_arg_gen_class float_gen;
-  float_gen.set_optarg (argstr);
-  flag_opt_arg_gen_class flag_gen;
-
+  update_arg_gen_class update_arg_gen;
+  update_arg_gen.set_optarg(argstr);
+  update_arg_gen.set_opt_var (opt->var_arg);
+  
   if (opt->multiple)
     {
-      int_gen.set_structure (string (opt->var_arg) + "_new");
-      str_gen_no_free.set_structure (string (opt->var_arg) + "_new");
-      long_gen.set_structure (string (opt->var_arg) + "_new");
-      long_long_gen.set_structure (string (opt->var_arg) + "_new");
-      float_gen.set_structure (string (opt->var_arg) + "_new");
+      update_arg_gen.set_structure (string (opt->var_arg) + "_new");
     }
   else
     {
-      int_gen.set_structure (ARGS_STRUCT);
-      str_gen.set_structure (ARGS_STRUCT);
-      long_gen.set_structure (ARGS_STRUCT);
-      long_long_gen.set_structure (ARGS_STRUCT);
-      float_gen.set_structure (ARGS_STRUCT);
+      update_arg_gen.set_structure (ARGS_STRUCT);
     }
 
   if (opt->arg_is_optional)
     {
       if (opt->type == ARG_STRING && opt->multiple)
         {
-          string_opt_arg_no_free_gen_class str_gen_null;
+          update_arg_gen_class str_gen_null;
+          str_gen_null.set_string_arg(true);
+          str_gen_null.set_no_free(true);
           str_gen_null.set_optarg ("NULL");
           str_gen_null.set_structure (string (opt->var_arg) + "_new");
           str_gen_null.set_opt_var (opt->var_arg);
-          str_gen_null.generate_string_opt_arg_no_free (stream, indent);
+          str_gen_null.generate_update_arg (stream, indent);
           stream << endl;
         }
 
@@ -272,58 +256,43 @@ CmdlineParserCreator::do_update_arg (struct gengetopt_option *opt,
 
   switch (opt->type)
     {
-    case ARG_NO:
-      stream << "break;";
-      break;
     case ARG_FLAG:
-      flag_gen.set_opt_var (opt->var_arg);
-      flag_gen.generate_flag_opt_arg (stream, indent);
+      update_arg_gen.set_flag_arg(true);
       break;
     case ARG_STRING:
+      update_arg_gen.set_string_arg(true);
       if (opt->multiple)
         {
-          str_gen_no_free.set_opt_var (opt->var_arg);
-          str_gen_no_free.generate_string_opt_arg_no_free (stream, indent);
+          update_arg_gen.set_no_free(true);
         }
       else
         {
-          str_gen.set_opt_var (opt->var_arg);
-          str_gen.generate_string_opt_arg (stream, indent);
+          update_arg_gen.set_no_free(false);
         }
       break;
     case ARG_INT:
-      int_gen.set_cast ("");
-      int_gen.set_opt_var (opt->var_arg);
-      int_gen.generate_int_opt_arg (stream, indent);
+      update_arg_gen.set_int_arg(true);
       break;
     case ARG_SHORT:
-      int_gen.set_cast ("(short)");
-      int_gen.set_opt_var (opt->var_arg);
-      int_gen.generate_int_opt_arg (stream, indent);
+      update_arg_gen.set_int_arg(true);
+      update_arg_gen.set_cast ("(short)");
       break;
     case ARG_LONG:
-      long_gen.set_cast ("");
-      long_gen.set_opt_var (opt->var_arg);
-      long_gen.generate_long_opt_arg (stream, indent);
+      update_arg_gen.set_long_arg(true);
       break;
     case ARG_FLOAT:
-      float_gen.set_cast ("(float)");
-      float_gen.set_opt_var (opt->var_arg);
-      float_gen.generate_float_opt_arg (stream, indent);
+      update_arg_gen.set_float_arg(true);
+      update_arg_gen.set_cast ("(float)");
       break;
     case ARG_DOUBLE:
-      float_gen.set_cast ("");
-      float_gen.set_opt_var (opt->var_arg);
-      float_gen.generate_float_opt_arg (stream, indent);
+      update_arg_gen.set_float_arg(true);
       break;
     case ARG_LONGDOUBLE:
-      float_gen.set_cast ("(long double)");
-      float_gen.set_opt_var (opt->var_arg);
-      float_gen.generate_float_opt_arg (stream, indent);
+      update_arg_gen.set_float_arg(true);
+      update_arg_gen.set_cast ("(long double)");
       break;
     case ARG_LONGLONG:
-      long_long_gen.set_opt_var (opt->var_arg);
-      long_long_gen.generate_long_long_opt_arg (stream, indent);
+      update_arg_gen.set_longlong_arg(true);
       break;
     default:
       fprintf (stderr, "gengetopt: bug found in %s:%d\n", __FILE__,
@@ -331,6 +300,8 @@ CmdlineParserCreator::do_update_arg (struct gengetopt_option *opt,
       abort ();
     }
 
+  update_arg_gen.generate_update_arg (stream, indent);
+    
   if (opt->arg_is_optional)
     {
       stream << endl;
@@ -354,7 +325,7 @@ CmdlineParserCreator::generate ()
 int
 CmdlineParserCreator::generate_header_file ()
 {
-  if (gengetopt_options == 0)
+  if (! gengetopt_options.size())
     {
       fprintf (stderr, "gengetopt: none option given\n");
       return 1;
@@ -377,8 +348,6 @@ CmdlineParserCreator::generate_option_arg(ostream &stream,
                                           unsigned int indent)
 {
   struct gengetopt_option * opt;
-  string indent_str (indent, ' ');
-  bool first = true;
 
   foropt
     if (opt->type != ARG_NO) {
@@ -392,11 +361,6 @@ CmdlineParserCreator::generate_option_arg(ostream &stream,
       case ARG_DOUBLE:
       case ARG_LONGDOUBLE:
       case ARG_LONGLONG:
-        if (! first)
-          stream << indent_str;
-        else
-          first = false;
-
         _generate_option_arg (stream, indent, opt);
 
         break;
@@ -413,52 +377,36 @@ _generate_option_arg(ostream &stream,
                      unsigned int indent,
                      struct gengetopt_option *opt)
 {
-  ostringstream opt_stream;
+  option_arg_gen_class option_arg_gen;
 
+  string type = arg_types[opt->type];
   if (opt->multiple)
-    opt_stream << "* ";
-  else
-    opt_stream << " ";
-
-  if (opt->type == ARG_FLAG)
-    opt_stream << opt->var_arg << "_flag";
-  else
-    opt_stream << opt->var_arg << "_arg";
-
-  opt_stream << ";\t/* " << opt->desc;
+    type += "*";
+  
+  option_arg_gen.set_type(type);
+  option_arg_gen.set_flag_arg((opt->type == ARG_FLAG));
+  option_arg_gen.set_desc(opt->desc);
+  option_arg_gen.set_name(opt->var_arg);
 
   if (opt->default_given)
     {
-      opt_stream << " (default='" << opt->default_string << "')";
+      option_arg_gen.set_has_default(true);
+      option_arg_gen.set_default_value(opt->default_string);
     }
 
   if (opt->type == ARG_FLAG)
     {
-      if (opt->flagstat)
-        opt_stream << " (default=on)";
-      else
-        opt_stream << " (default=off)";
+      option_arg_gen.set_default_on(opt->flagstat);
     }
 
-  opt_stream << ".  */\n";
-  const string opt_string = opt_stream.str ();
-
   if (opt->type == ARG_LONGLONG)
     {
-      string indent_str (indent, ' ');
-
-      stream << "#ifdef HAVE_LONG_LONG\n";
-      stream << indent_str;
-      stream << arg_types[opt->type] << opt_string;
-      stream << indent_str;
-      stream << "#else\n";
-      stream << indent_str;
-      stream << arg_types[ARG_LONG] << opt_string;
-      stream << indent_str;
-      stream << "#endif\n";
+      option_arg_gen.set_long_long_arg(true);
+      option_arg_gen.set_longlongtype(arg_types[opt->type]);
+      option_arg_gen.set_longtype(arg_types[ARG_LONG]);
     }
-  else
-    stream << arg_types[opt->type] << opt_string;
+
+  option_arg_gen.generate_option_arg(stream, indent);
 }
 
 void
@@ -468,6 +416,7 @@ CmdlineParserCreator::generate_option_given(ostream &stream,
   struct gengetopt_option * opt;
   string indent_str (indent, ' ');
   bool first = true;
+  given_field_gen_class given_gen;
 
   foropt
     {
@@ -492,11 +441,11 @@ CmdlineParserCreator::generate_option_given(ostream &stream,
       else
         first = false;
 
-      if (opt->multiple)
-        stream << "unsigned ";
-
-      stream << "int " << opt->var_arg <<
-        "_given ;\t/* Whether " << opt->long_opt << " was given.  */\n";
+      given_gen.set_multiple (opt->multiple);
+      given_gen.set_arg_name (opt->var_arg);
+      given_gen.set_long_opt (opt->long_opt);
+      given_gen.set_group (opt->multiple && opt->group_value);
+      given_gen.generate_given_field (stream);
     }
 
   if (unamed_options)
@@ -510,36 +459,66 @@ CmdlineParserCreator::generate_option_given(ostream &stream,
 }
 
 void
-CmdlineParserCreator::generate_print_purpose(ostream &stream,
-                                             unsigned int indent)
+CmdlineParserCreator::generate_option_values_decl(ostream &stream,
+                                                  unsigned int indent)
 {
-  if (gengetopt_purpose != NULL)
+  struct gengetopt_option * opt;
+  bool first = true;
+
+  foropt
     {
-      string indent_str (indent, ' ');
+      if (opt->acceptedvalues) {
+        if (first) {
+          first = false;
+        }
 
-      string wrapped_purpose;
-      wrap_cstr(wrapped_purpose, 2, 0, gengetopt_purpose);
-
-      stream << indent_str;
-      stream << "\"Purpose:\\n\"";
-      stream << endl;
-      stream << "\"  ";
-      stream << wrapped_purpose;
-      stream << "\\n\""; // close the last line
-      stream << endl;
-      stream << indent_str;
-      stream << "\"\\n\"\n"; // generate the last \n
+        stream << "extern char *" << opt->var_arg <<
+          "_values[] ;\t/* Possible values for " << opt->long_opt << ".  */\n";
+      }
     }
+
+  if (! first)
+    stream << "\n";
 }
 
-string
-CmdlineParserCreator::generate_usage_string()
+void
+CmdlineParserCreator::generate_option_values(ostream &stream,
+                                             unsigned int indent)
+{
+  struct gengetopt_option * opt;
+  bool first = true;
+
+  foropt
+    {
+      if (opt->acceptedvalues) {
+        if (first) {
+          first = false;
+        }
+
+        stream << "char *" << opt->var_arg <<
+          "_values[] = {" << opt->acceptedvalues->toString(false) <<
+          ", 0} ;\t/* Possible values for " << opt->long_opt << ".  */\n";
+      }
+    }
+
+  if (! first)
+    stream << "\n";
+}
+
+const string
+CmdlineParserCreator::generate_usage_string(bool use_config_package)
 {
   struct gengetopt_option * opt;
   ostringstream usage;
   const char   *type_str;
 
-  usage << "Usage: %s ";
+  usage << "Usage: ";
+  if (gengetopt_package)
+    usage << gengetopt_package;
+  else if (use_config_package)
+    usage << "%s";
+    
+  usage << " ";
 
   if ( long_help )
     {
@@ -626,15 +605,28 @@ CmdlineParserCreator::generate_usage_string()
 static void
 generate_help_desc_print(ostream &stream,
                          unsigned int desc_column,
-                         const char *descript, const char *defval)
+                         const char *descript, const char *defval,
+                         const string &values)
 {
   string desc;
   string desc_with_default = descript;
 
-  if (defval)
+  if (defval || values.size())
     {
-      desc_with_default += "  (default=";
-      desc_with_default += defval;
+      desc_with_default += "  (";
+
+      if (values.size()) {
+        desc_with_default += "possible values=";
+        desc_with_default += values;
+        if (defval)
+          desc_with_default += " ";
+      }
+
+      if (defval) {
+        desc_with_default += "default=";
+        desc_with_default += defval;
+      }
+
       desc_with_default += ")";
     }
 
@@ -648,12 +640,44 @@ void
 CmdlineParserCreator::generate_help_option_print(ostream &stream,
                                                  unsigned int indent)
 {
+  OptionHelpList *option_list = generate_help_option_list();
+  
+  print_help_string_gen_class print_gen;
+  
+  for (OptionHelpList::const_iterator it = option_list->begin();
+       it != option_list->end(); ++it)
+    {
+      print_gen.set_helpstring(*it);
+      print_gen.generate_print_help_string(stream, indent);
+    }
+  
+  delete option_list;
+}
+
+const string
+CmdlineParserCreator::generate_purpose()
+{
+  string wrapped_purpose;
+  
+  if (gengetopt_purpose != NULL)
+    {
+      wrap_cstr(wrapped_purpose, 0, 0, gengetopt_purpose);
+    }
+    
+  return wrapped_purpose;
+}
+
+OptionHelpList *
+CmdlineParserCreator::generate_help_option_list()
+{
+  OptionHelpList *option_list = new OptionHelpList;
+  
   long desc_col;
   struct gengetopt_option * opt;
-  string indent_str (indent, ' ');
-
-  int                type_len;
+  
+  int           type_len;
   const char   *type_str;
+  ostringstream stream;
 
   /* calculate columns */
   desc_col = 0;
@@ -689,11 +713,6 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
   char *curr_section = 0;
   bool first_option = true;
 
-  stream << indent_str;
-  stream << "printf(\"\\n\");\n";
-
-  print_help_string_gen_class print_gen;
-
   foropt
     {
       if (opt->group_value &&
@@ -708,12 +727,10 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
               wrap_cstr (wrapped_desc, 2, 0, opt->group_desc);
             }
 
-          stream << indent_str;
           group_string += opt->group_value + wrapped_desc;
-          print_gen.set_helpstring (group_string);
-          print_gen.generate_print_help_string (stream, indent);
-          stream << endl;
-
+          
+          option_list->push_back (group_string);
+          
           prev_group = opt->group_value;
         }
 
@@ -728,20 +745,17 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
             sec_string << "\\n";
 
           sec_string << opt->section << ":" ;
-          print_gen.set_helpstring (sec_string.str());
-          stream << indent_str;
-          print_gen.generate_print_help_string (stream, indent);
-          stream << endl;
-
+          
+          string wrapped_def;
+          wrap_cstr(wrapped_def, 0, 0, sec_string.str());
+          option_list->push_back(wrapped_def);
+          
           if (opt->section_desc)
             {
               string wrapped_desc ( 2, ' ');
               wrap_cstr ( wrapped_desc, 2, 0, opt->section_desc );
 
-              stream << indent_str;
-              print_gen.set_helpstring ( wrapped_desc );
-              print_gen.generate_print_help_string (stream, indent);
-              stream << endl;
+              option_list->push_back(wrapped_desc);
             }
         }
 
@@ -749,9 +763,6 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
       const char * def_val = NULL;
       string def_str = "`";
 
-      stream << indent_str;
-      stream << "printf(\"%s\\n\",\"";
-
       ostringstream option_stream;
 
       if (opt->type == ARG_FLAG || opt->type == ARG_NO)
@@ -817,9 +828,14 @@ CmdlineParserCreator::generate_help_option_print(ostream &stream,
           stream << indent;
         }
 
-      generate_help_desc_print(stream, desc_col, opt_desc, def_val);
-      stream << "\");\n";
+      generate_help_desc_print(stream, desc_col, opt_desc, def_val,
+        (opt->acceptedvalues ? opt->acceptedvalues->toString() : ""));
+      
+      option_list->push_back(stream.str());
+      stream.str("");
     }
+    
+  return option_list;
 }
 
 void
@@ -828,13 +844,16 @@ CmdlineParserCreator::generate_given_init(ostream &stream,
 {
   struct gengetopt_option * opt;
   string indent_str (indent, ' ');
+  clear_given_gen_class clear_given;
+  clear_given.set_arg_struct(ARGS_STRUCT);
 
   /* now we initialize "given" fields */
   foropt
     {
       stream << indent_str;
-      stream << ARGS_STRUCT << "->" << opt->var_arg << "_given = 0 ;";
-      stream << endl;
+      clear_given.set_var_arg(opt->var_arg);
+      clear_given.set_group(opt->multiple && opt->group_value);
+      clear_given.generate_clear_given(stream);
     }
 
   groups_collection_t::const_iterator end = gengetopt_groups.end();
@@ -854,6 +873,8 @@ CmdlineParserCreator::generate_reset_groups(ostream &stream, unsigned int indent
   string indent_str (indent, ' ');
   ostringstream body;
   reset_group_gen_class reset_group;
+  clear_given_gen_class clear_given;
+  clear_given.set_arg_struct(ARGS_STRUCT);
 
   reset_group.set_args_info (c_source_gen_class::args_info);
 
@@ -863,15 +884,19 @@ CmdlineParserCreator::generate_reset_groups(ostream &stream, unsigned int indent
     {
       body.str ("");
       bool found_option = false;
+      bool multiple_arg = false;
 
       foropt
       {
         if (opt->group_value && strcmp(opt->group_value, idx->first.c_str()) == 0)
           {
             /* now we reset "given" fields */
-            body << indent_str;
-            body << ARGS_STRUCT << "->" << opt->var_arg << "_given = 0 ;";
-            body << endl;
+            stream << indent_str;
+            clear_given.set_var_arg(opt->var_arg);
+            if (opt->multiple && opt->group_value)
+              multiple_arg = true;
+            clear_given.set_group(opt->multiple && opt->group_value);
+            clear_given.generate_clear_given(body);
 
             free_option (opt, body, indent);
             found_option = true;
@@ -880,6 +905,7 @@ CmdlineParserCreator::generate_reset_groups(ostream &stream, unsigned int indent
 
       if (found_option)
         {
+          reset_group.set_multiple_arg(multiple_arg);
           reset_group.set_name (canonize_name (idx->first));
           reset_group.set_body (body.str ());
           reset_group.generate_reset_group (stream);
@@ -960,15 +986,16 @@ CmdlineParserCreator::generate_list_def(ostream &stream, unsigned int indent)
     {
       if (opt->multiple)
         {
-          if (! generated_counter)
-            {
-              stream << indent_str;
-              stream << "int i;        /* Counter */" << endl;
-              generated_counter = true;
-              stream << endl;
-            }
           if (opt->type)
             {
+              if (! generated_counter)
+              {
+                stream << indent_str;
+                stream << "int i;        /* Counter */" << endl;
+                generated_counter = true;
+                stream << endl;
+              }
+          
               stream << indent_str;
               multiple_opt_list.set_arg_name (opt->var_arg);
               multiple_opt_list.generate_multiple_opt_list (stream, indent);
@@ -1165,7 +1192,8 @@ CmdlineParserCreator::generate_handle_help(ostream &stream,
      help_gen.set_short_option (HELP_SHORT_OPT_STR);
      help_gen.set_option_comment (HELP_OPT_DESCR);
      help_gen.set_option_var_name (HELP_LONG_OPT);
-     help_gen.set_package_var_name (c_source_gen_class::package_var_name);
+     help_gen.set_package_var_name (EXE_NAME);
+     help_gen.set_has_short_option (true);
 
      help_gen.generate_generic_option (stream, indent);
 
@@ -1191,7 +1219,8 @@ CmdlineParserCreator::generate_handle_version(ostream &stream,
      version_gen.set_short_option (VERSION_SHORT_OPT_STR);
      version_gen.set_option_comment (VERSION_OPT_DESCR);
      version_gen.set_option_var_name (VERSION_LONG_OPT);
-     version_gen.set_package_var_name (c_source_gen_class::package_var_name);
+     version_gen.set_package_var_name (EXE_NAME);
+     version_gen.set_has_short_option (true);
 
      version_gen.generate_generic_option (stream, indent);
 
@@ -1210,152 +1239,78 @@ void
 CmdlineParserCreator::generate_handle_no_short_option(ostream &stream,
                                                       unsigned int indent)
 {
-  const string tmpvar("multi_token"); // reuse the `c' variable (returning getopt_long value)
-  const string optarg("optarg");
-
-  struct gengetopt_option * opt;
-  no_short_option_gen_class opt_gen;
-  string indent_str (indent, ' ');
-  opt_gen.set_package_var_name (c_source_gen_class::package_var_name);
-  multiple_option_no_short_gen_class multip_opt_gen;
-  multiple_option_new_list_node_gen_class multi_gen_node;
-  bool first = true;
-
-  foropt
-    if (! opt->short_opt)
-      {
-        stream << indent_str;
-        ostringstream str_stream;
-
-        if (opt->multiple)
-          {
-            ostringstream list_stream;
-
-            multip_opt_gen.set_option_comment (opt->desc);
-            multip_opt_gen.set_long_option (opt->long_opt);
-            multip_opt_gen.set_option_var_name (opt->var_arg);
-
-            if (opt->type)
-              { // arguments possibly exists
-                do_update_arg (opt, str_stream, 0, tmpvar);
-
-                multi_gen_node.set_optarg(optarg);
-                multi_gen_node.set_option_var_name (opt->var_arg);
-                multi_gen_node.set_update_arg(str_stream.str());
-                multi_gen_node.generate_multiple_option_new_list_node (list_stream);
-              }
-
-            multip_opt_gen.set_update_arg ("break;");
-
-            multip_opt_gen.set_new_node_list (list_stream.str ());
-
-            generic_option_group_gen_class group_gen;
-            if (opt->group_value)
-              {
-                str_stream.str ("");
-                str_stream << endl;
-                str_stream << "  ";
-                group_gen.set_group_var_name
-                  (canonize_name (opt->group_value));
-                group_gen.generate_generic_option_group (str_stream, 2);
-                multip_opt_gen.set_update_group_count (str_stream.str ());
-              }
-            else
-              multip_opt_gen.set_update_group_count ("");
-
-            multip_opt_gen.generate_multiple_option_no_short
-              (stream, indent);
-          }
-        else
-          {
-            opt_gen.set_option_comment (opt->desc);
-            opt_gen.set_long_option (opt->long_opt);
-            opt_gen.set_option_var_name (opt->var_arg);
-
-            do_update_arg (opt, str_stream, 0);
-            opt_gen.set_update_arg (str_stream.str ());
-
-            generic_option_group_gen_class group_gen;
-            if (opt->group_value)
-              {
-                str_stream.str ("");
-                str_stream << endl;
-                str_stream << "  ";
-                group_gen.set_group_var_name
-                  (canonize_name (opt->group_value));
-                group_gen.generate_generic_option_group (str_stream, 2);
-                opt_gen.set_update_group_count (str_stream.str ());
-              }
-            else
-              opt_gen.set_update_group_count ("");
-
-            opt_gen.generate_no_short_option (stream, indent);
-          }
-
-        stream << endl;
-
-        if (first)
-          {
-            first = false;
-            opt_gen.set_gen_else ("else ");
-            multip_opt_gen.set_gen_else ("else ");
-          }
-      }
-
-  if (! first) // something has been generated
-    {
-      generateBreak(stream, indent);
-    }
-
-  stream << endl;
+  handle_options(stream, indent, false);
 }
 
 void
 CmdlineParserCreator::generate_handle_option(ostream &stream,
                                              unsigned int indent)
+{
+  handle_options(stream, indent, true);
+}
+
+void
+CmdlineParserCreator::handle_options(ostream &stream, unsigned int indent, bool has_short)
 {
   const string tmpvar("multi_token"); // reuse the `c' variable (returning getopt_long value)
   const string optarg("optarg");
 
   struct gengetopt_option * opt;
   generic_option_gen_class option_gen;
-  multiple_option_gen_class multi_gen;
-  multiple_option_new_list_node_gen_class multi_gen_node;
+  multiple_option_gen_class multip_opt_gen;
   string indent_str (indent, ' ');
+  bool first = true;
 
-  option_gen.set_package_var_name (c_source_gen_class::package_var_name);
+  option_gen.set_package_var_name (EXE_NAME);
+  option_gen.set_has_short_option (has_short);
+  multip_opt_gen.set_package_var_name (EXE_NAME);
+  multip_opt_gen.set_has_short_option (has_short);
 
   foropt
     {
-      if (opt->short_opt)
+      if ((has_short && opt->short_opt) || (!has_short && !opt->short_opt))
         {
           if (opt->short_opt == 'h' || opt->short_opt == 'V')
             continue;
 
-          stream << indent_str;
+          if (has_short || first)
+            stream << indent_str;
 
           string short_opt (1, opt->short_opt);
           if (opt->multiple)
             {
-              multi_gen.set_short_option (short_opt);
-              multi_gen.set_option_comment (opt->desc);
-              multi_gen.set_option_var_name (opt->var_arg);
+              ostringstream list_stream;
+              ostringstream str_stream;
 
-              multi_gen.generate_multiple_option (stream, indent);
+              multip_opt_gen.set_short_option (short_opt);
+              multip_opt_gen.set_option_comment (opt->desc);
+              multip_opt_gen.set_long_option (opt->long_opt);
+              multip_opt_gen.set_option_var_name (opt->var_arg);
+              multip_opt_gen.set_option_has_values (opt->acceptedvalues != 0);
 
               if (opt->type)
-                {
-                  ostringstream update_arg;
+                { // arguments possibly exists
+                  do_update_arg (opt, str_stream, 0, tmpvar);
+                  multip_opt_gen.set_option_has_type(true);
 
-                  do_update_arg (opt, update_arg, 0, tmpvar);
-
-                  multi_gen_node.set_optarg(optarg);
-                  multi_gen_node.set_option_var_name (opt->var_arg);
-                  multi_gen_node.set_update_arg(update_arg.str());
-                  multi_gen_node.generate_multiple_option_new_list_node (stream, indent+2);
+                  multip_opt_gen.set_update_arg(str_stream.str());
                 }
               else
-                do_update_arg (opt, stream, indent + 2);
+                multip_opt_gen.set_option_has_type(false);
+              
+              generic_option_group_gen_class group_gen;
+              if (opt->group_value)
+                {
+                  multip_opt_gen.set_group_var_name (canonize_name (opt->group_value));
+                  multip_opt_gen.set_option_has_group(true);                  
+                }
+              else
+                multip_opt_gen.set_option_has_group(false);
+
+              multip_opt_gen.set_option_not_string(opt->type != ARG_STRING);
+              // for strings multi_token is copied directly, so it must not be freed.
+              
+              multip_opt_gen.generate_multiple_option (stream, indent);
             }
           else
             {
@@ -1363,29 +1318,55 @@ CmdlineParserCreator::generate_handle_option(ostream &stream,
               option_gen.set_option_comment (opt->desc);
               option_gen.set_long_option (opt->long_opt);
               option_gen.set_option_var_name (opt->var_arg);
+              option_gen.set_option_has_values (opt->acceptedvalues != 0);
 
-              option_gen.generate_generic_option (stream, indent);
-
-              // NOTE: I don't know why the following code was shared with the branch for
-              // multiple options because you can't specify multiple groupoption
+              ostringstream update_stream;
+              
               generic_option_group_gen_class group_gen;
               if (opt->group_value)
                 {
+                  ostringstream group_stream;
+                  
+                  group_stream << endl;
+                  group_stream << "  ";
                   group_gen.set_group_var_name
                     (canonize_name (opt->group_value));
-                  group_gen.generate_generic_option_group (stream, indent+2);
+                  group_gen.generate_generic_option_group (group_stream, 2);
+                  option_gen.set_update_group_count (group_stream.str ());
                 }
+              else
+                option_gen.set_update_group_count ("");
 
-              do_update_arg (opt, stream, indent + 2);
+              do_update_arg (opt, update_stream, 0);
+              
+              option_gen.set_update_arg(update_stream.str());
+              
+              option_gen.generate_generic_option (stream, indent);
+              
+              if (opt->type != ARG_NO && has_short)
+                generateBreak(stream, indent + 2);
             }
 
-          if (opt->type != ARG_NO)
-            generateBreak(stream, indent + 2);
-
-          stream << endl;
-          stream << endl;
+          if (has_short)
+            {
+              stream << endl;
+              stream << endl;
+            }
+          
+          if (first && !has_short)
+            {
+              first = false;
+              option_gen.set_gen_else ("else ");
+              multip_opt_gen.set_gen_else ("else ");
+            }
         }
     }
+    
+  if (! first && !has_short) // something has been generated
+    {
+      generateBreak(stream, indent);
+      stream << endl;
+    }
 }
 
 #define GROUP_REQUIRED_COMPARISON "!="
@@ -1399,7 +1380,7 @@ CmdlineParserCreator::generate_handle_group(ostream &stream,
 {
   group_option_gen_class opt_gen;
   string indent_str (indent, ' ');
-  opt_gen.set_package_var_name (c_source_gen_class::package_var_name);
+  opt_gen.set_package_var_name (EXE_NAME);
 
   groups_collection_t::const_iterator end = gengetopt_groups.end();
   for ( groups_collection_t::const_iterator idx = gengetopt_groups.begin();
@@ -1430,7 +1411,7 @@ CmdlineParserCreator::generate_handle_required(ostream &stream,
 {
   struct gengetopt_option * opt;
   required_option_gen_class opt_gen;
-  opt_gen.set_package_var_name (c_source_gen_class::package_var_name);
+  opt_gen.set_package_var_name ("prog_name");
   string indent_str (indent, ' ');
 
   /* write test for required options */
@@ -1522,4 +1503,24 @@ CmdlineParserCreator::generate_free(ostream &stream,
     }
 }
 
+void
+CmdlineParserCreator::generate_list_free(ostream &stream,
+                                         unsigned int indent)
+{
+  struct gengetopt_option * opt;
+
+  if (! has_multiple_options())
+    return;
+  
+  free_list_gen_class free_list;
+  
+  foropt
+    {
+      if (opt->multiple && opt->type) {
+        free_list.set_list_name(opt->var_arg);
+        free_list.set_string_list(opt->type == ARG_STRING);
+        free_list.generate_free_list(stream, indent);
+      }
+    }
+}
 
