@@ -48,10 +48,11 @@
      TOK_TYPESTR = 274,
      TOK_SECTION = 275,
      TOK_SECTIONDESC = 276,
-     TOK_STRING = 277,
-     TOK_MLSTRING = 278,
-     TOK_CHAR = 279,
-     TOK_ARGTYPE = 280
+     TOK_VALUES = 277,
+     TOK_STRING = 278,
+     TOK_MLSTRING = 279,
+     TOK_CHAR = 280,
+     TOK_ARGTYPE = 281
    };
 #endif
 #define TOK_PACKAGE 258
@@ -73,24 +74,26 @@
 #define TOK_TYPESTR 274
 #define TOK_SECTION 275
 #define TOK_SECTIONDESC 276
-#define TOK_STRING 277
-#define TOK_MLSTRING 278
-#define TOK_CHAR 279
-#define TOK_ARGTYPE 280
+#define TOK_VALUES 277
+#define TOK_STRING 278
+#define TOK_MLSTRING 279
+#define TOK_CHAR 280
+#define TOK_ARGTYPE 281
 
 
 
 
 #if ! defined (YYSTYPE) && ! defined (YYSTYPE_IS_DECLARED)
-#line 63 "parser.yy"
+#line 68 "parser.yy"
 typedef union YYSTYPE {
     char   *str;
     char    chr;
     int	    argtype;
     int	    boolean;
+    class AcceptedValues *ValueList;
 } YYSTYPE;
 /* Line 1240 of yacc.c.  */
-#line 93 "parser.h"
+#line 96 "parser.h"
 # define yystype YYSTYPE /* obsolescent; will be withdrawn */
 # define YYSTYPE_IS_DECLARED 1
 # define YYSTYPE_IS_TRIVIAL 1
